from pathlib import Path

def isComment(line):
    return ';' in line or '[' in line

def removeComments(tokens):
    index = next((i for i, e in enumerate(tokens) if isComment(e)), len(tokens))
    return tokens[0:index]

class MemoryRegion():
    def __init__(self, start, end):
        self.start = start
        self.end = end
        self.size = self.end - self.start + 4

    def __str__(self):
        return hex(self.start) + " - " + hex(self.end)

    def __repr__(self):
        return str(self)

class ConfigParser():
    def __init__(self, file):
        self.sourceFiles = []
        self.includePaths = []
        self.compilerFlags = []
        self.linkerFlags = []
        self.libraries = []
        self.staticOverwrites = {}
        self.fixedSymbols = {}
        self.memoryRegions = []
        self.basedir = str(Path(file).parents[0]) + '/'

        # tokenize file, and identify '=' locations to determine variables
        with open(file, 'r') as f:
            lines = [line.split() for line in f]
        lines = [removeComments(line) for line in lines]
        lines = [x for x in lines if x]
        tokens = [token for line in lines for token in line]
        breaks = [i - 1 for i, e in enumerate(tokens) if e == '=']

        switcher = {
            'REGIONS': self.processMemoryRegions,
            'SOURCES': self.processSourceFiles,
            'LIBRARIES': self.processLibraries,
            'INCLUDE_PATHS': self.processIncludes,
            'COMPILER_FLAGS': self.processCompilerFlags,
            'LINKER_FLAGS': self.processLinkerFlags,
            'FIXED_SYMBOLS': self.processFixedSymbols
        }
        for start, stop in zip(breaks, breaks[1:] + [len(tokens)]):
            var = tokens[start:stop]
            func = switcher.get(var[0], self.processStaticOverwrite)
            func(var[0], var[2:])

    def processMemoryRegions(self, name, var): #pylint: disable=unused-argument
        for region in var:
            start = int(region.split('-')[0], 16)
            end = int(region.split('-')[1], 16)
            self.memoryRegions.append(MemoryRegion(start, end))

    def processSourceFiles(self, name, var): #pylint: disable=unused-argument
        self.sourceFiles = var

    def processLibraries(self, name, var): #pylint: disable=unused-argument
        self.libraries = [self.basedir + x for x in var]

    def processIncludes(self, name, var): #pylint: disable=unused-argument
        self.includePaths = [self.basedir + x for x in var]

    def processCompilerFlags(self, name, var): #pylint: disable=unused-argument
        self.compilerFlags = var

    def processLinkerFlags(self, name, var): #pylint: disable=unused-argument
        self.linkerFlags = var

    def processFixedSymbols(self, name, var): #pylint: disable=unused-argument
        for i in range(0, len(var), 3):
            self.fixedSymbols[var[i]] = [var[i+1], var[i+2]]

    def processStaticOverwrite(self, name, var): #pylint: disable=unused-argument
        self.staticOverwrites[int('0x' + name, 0)] = int('0x' + var[0], 0)

    def getStaticOverwrites(self):
        return self.staticOverwrites

    def getZeroedMemory(self):
        code = {}
        for region in self.memoryRegions:
            for address in range(region.start, region.end, 4):
                code[address] = 0
        return code

    def print(self):
        print("\n======== Configuration ================\n")
        print("Compiler: powerpc-eabi-gcc")
        print("Flags:", ' '.join(self.compilerFlags))
        print("Linker: powerpc-eabi-ld")
        print("Flags:", ' '.join(self.linkerFlags))
        print("Include Paths:")
        print(*self.includePaths, sep='\n')
        print("Libraries:")
        print(*self.libraries, sep='\n')
        print("\n======== Available Memory Regions =====\n")
        print(*self.memoryRegions, sep='\n')
        print("\n======== Static Overwrites ============\n")
        for key, val in self.staticOverwrites.items():
            print(hex(val) + ' at address ' + hex(key))
        print("\n======== Code Entry Points ============\n")
        for key, val in self.fixedSymbols.items():
            print(key + '() at address 0x' + val[0])
        print("\n======== Source Files =================\n")
        print(*self.sourceFiles, sep='\n')
        print("\n=======================================\n")



import sys
import hashlib
from bisect import bisect
import struct

def readFile(file, pos, size=4):
    with open(file, 'rb') as f:
        f.seek(pos)
        return struct.unpack(">I", f.read(size))[0]

def writeFile(file, val, pos, size=4):
    val = val.to_bytes(size, byteorder='big')
    with open(file, 'r+b') as f:
        f.seek(pos)
        f.write(val)

class DolSection():
    def __init__(self, name, filePos, address, size):
        self.name = name
        self.filePos = filePos
        self.address = address
        self.size = size

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

class DolTable():
    def __init__(self, file):
        # get table information
        self.tablePos = readFile(file, 0x420)
        self.bssAddress = readFile(file, self.tablePos + 0xd8)
        self.bssSize = readFile(file, self.tablePos + 0xdc)
        self.entryPoint = readFile(file, self.tablePos + 0xe0)

        # process sections
        self.unsortedSections = []
        names = ["Text" + str(i) for i in range(7)]
        names += ["Data" + str(i) for i in range(11)]
        for i, name in enumerate(names):
            filePos = readFile(file, self.tablePos + i * 4)
            address = readFile(file, self.tablePos + i * 4 + 0x48)
            size = readFile(file, self.tablePos + i * 4 + 0x90)
            self.unsortedSections.append(DolSection(name, filePos, address, size))
        self.sections = sorted(self.unsortedSections, key=lambda x: x.address)

        # calculate start and end of code
        self.codeStart = min([x.address for x in self.sections if x.address > 0])
        self.codeEnd = self.sections[-1].address + self.sections[-1].size

    def __str__(self):
        lines = ["Code Start: " + hex(self.codeStart)]
        lines.append("Code End: " + hex(self.codeEnd))
        lines.append("Entry Point: " + hex(self.entryPoint))
        lines.append("BSS Address: " + hex(self.bssAddress))
        lines.append("BSS Size: " + hex(self.bssSize))
        f = '{0:>7} {1:>14} {2:>15} {3:>13}'
        lines.append(f.format('Section', 'File Position', 'Memory Address',
                              'Section Size'))
        for x in self.unsortedSections:
            lines.append(f.format(x.name, hex(x.filePos), hex(x.address), hex(x.size)))
        return '\n'.join(lines)

    def __repr__(self):
        return str(self)

    def filePos(self, address, verbose=False):
        if self.codeStart <= address <= self.codeEnd:
            if verbose:
                print("interpreting", hex(address), "as a memory address")
            index = bisect([x.address for x in self.sections], address)
            section = self.sections[index-1]
            return self.tablePos + section.filePos + address - section.address
        return address # otherwise interpret as a pure file position

class Iso():
    def __init__(self, file):
        if readFile(file, 0x1c) != 0xc2339f3d: # magic number used to verify
            sys.exit("Error: not valid Gamecube/Wii ISO")
        self.file = file
        self.dolTable = DolTable(file)

    def read(self, address, verbose=False):
        pos = self.dolTable.filePos(address, verbose)
        return readFile(self.file, pos)

    def save(self, saveFile):
        with open(self.file, 'rb') as iFile, open(saveFile, 'wb') as sFile:
            pos = 0
            endPos = self.dolTable.filePos(self.dolTable.codeEnd)
            while pos <= endPos:
                byte = iFile.read(1)
                sFile.write(byte)
                pos += 1

    def load(self, saveFile):
        with open(self.file, 'r+b') as iFile, open(saveFile, 'rb') as sFile:
            iFile.seek(0)
            while True:
                byte = sFile.read(1)
                if byte:
                    iFile.write(byte)
                else:
                    break

    def bulkWrite(self, code):
        with open(self.file, 'r+b') as f:
            for addr, inst in code.items():
                inst = struct.pack(">I", inst)
                pos = self.dolTable.filePos(addr)
                f.seek(pos)
                f.write(inst)

    # computes md5 checksum
    # https://stackoverflow.com/a/3431838
    def md5(self):
        md5Hash = hashlib.md5()
        with open(self.file, "rb") as f:
            for chunk in iter(lambda: f.read(16384), b""):
                md5Hash.update(chunk)
        return md5Hash.hexdigest()



import sys
import argparse
from pathlib import Path
from wiimake.iso import Iso
from wiimake.config_parser import ConfigParser
from wiimake.pipeline import WiimakePipeline
from wiimake.pipeline import runCmd
from wiimake._version import __version__

# description for wiimake
DESCRIPTION = '''
wiimake is an automated tool for compiling, allocating, linking, and injecting
C code into games for the Gamecube/Wii. The user must provide a configuration
file that contains information about the sources and libraries that should be
injected, entry points of their code (function names), and which regions of RAM
are available for overwriting. Given this information, the program proceeds as
follows: (1) compile all source files (2) find sizes of each section of code
that needs to be injected (3) find an arrangement of the code that fits within
the memory regions provided by the user in the config file (4) run the linker
(5) inject the code into the appropiate addresses. In one command the user can
take their raw C files and produce a working iso.
'''

def main():
    """ main function for wiimake """
    print("WiiMake version:", __version__)

    # check for powerpc toolchain
    try:
        res = runCmd(['powerpc-eabi-gcc', '--version'])
    except FileNotFoundError:
        sys.exit("Error: Can't find devkitPPC toolchain")
    else:
        line = res.stdout.split('\n')[0]
        devkitppcVersion = ' '.join(line.split()[1:])

    # parse command line arguments
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--version', action='version', version='%(prog)s ' +  __version__)
    parser.add_argument('iso_file', help='Gamecube/Wii .iso file')
    parser.add_argument('config_file', help='wiimake configuration file')
    parser.add_argument('--save-temps', action='store_true',
                        help='save temporary build files')
    parser.add_argument('--verbose', action='store_true',
                        help='provide more verbose output')
    args = parser.parse_args()

    # Startup Message
    print("devkitPPC version:", devkitppcVersion)
    print("Building Code from", Path(args.config_file).name)
    print("Injecting in", Path(args.iso_file).name)
    print("")

    # parse configuration file
    config = ConfigParser(args.config_file)
    if args.verbose:
        config.print()

    # run wiimake pipeline to generate code at available addresses in memory
    pipeline = WiimakePipeline(config)
    pipeline.run(args.verbose, args.save_temps)

    # inject code into iso file
    print("== Injecting Code into Game ==")
    iso = Iso(args.iso_file)
    iso.bulkWrite(config.getStaticOverwrites())
    iso.bulkWrite(config.getZeroedMemory())
    iso.bulkWrite(pipeline.getCode())
    print("lines of code injected:", len(pipeline.getCode()))



SIZE_LINKER_SCRIPT_SECTION = '''
wiimake_section_{num} :
{{
    {section}
}}
_sizeof_section_{num}=SIZEOF(wiimake_section_{num});
'''

FINAL_LINKER_SCRIPT_SECTION = '''
wiimake_section_{num} {address} : /* section size = {size} */
{{
    {section}
}}
'''

def getFinalLinkerScript(sections):
    script = "SECTIONS {\n"
    for i, sec in enumerate(sections):
        script += FINAL_LINKER_SCRIPT_SECTION.format(num=i+1, section=sec,
                                                     address=hex(sec.address),
                                                     size=hex(sec.size))
    script += "}\n"
    return script

def getSizeLinkerScript(sections):
    script = "SECTIONS {\n"
    for i, sec in enumerate(sections):
        script += SIZE_LINKER_SCRIPT_SECTION.format(num=i+1, section=sec)
    script += "}\n"
    return script



# TO-DO make deterministic

#### Notes ####

# This is an important note about link order that explains some bugs seen in the
# past. In particular, it is important to order the object files so library
# files come at the end of the linker script.
#
# Excerpt from "An Introduction to GCC: for the GNU Compilers gcc and g++"
# by Brian Gough
#
# On Unix-like systems, the traditional behavior of compilers and linkers is to
# search for external functions from left to right in the object files
# specified on the command line. This means that the object file which
# contains the definition of a function should appear after any files which
# call that function.
# In this case, the file 'hello_fn.o' containing the function hello should be
# specified after 'main.o' itself, since main calls hello:
#   $ gcc main.o hello_fn.o -o hello   (correct order)
# With some compilers or linkers the opposite ordering would result in an error,
#   $ cc hello_fn.o main.o -o hello    (incorrect order)
#   main.o: In function `main':
#   main.o(.text+0xf): undefined reference to `hello'
# because there is no object file containing hello after 'main.o'.
# Most current compilers and linkers will search all object files, regardless
# of order, but since not all compilers do this it is best to follow the
# convention of ordering object files from left to right.
# This is worth keeping in mind if you ever encounter unexpected problems with
# undefined references, and all the necessary object files appear to be present
# on the command line.

###############

import os
import sys
import copy
import subprocess
from pathlib import Path
from wiimake.assembly_code import PRESERVE_REGISTERS_ASSEMBLY
from wiimake.assembly_code import INJECTION_POINT_ASSEMBLY
from wiimake.assembly_code import BRANCH_TO_ENTRY_POINT_ASSEMBLY
from wiimake.linker_scripts import getSizeLinkerScript
from wiimake.linker_scripts import getFinalLinkerScript
from wiimake.iso import readFile
from wiimake.iso import writeFile

def createFile(content, fname):
    """ write a string to a file """
    file = open(fname, 'w')
    file.write(content)
    file.close()

def runCmd(cmd, verbose=False, text=True):
    """ run a shell command as a subprocess """
    if verbose:
        print(''.join([x + ' ' for x in cmd]))
    return subprocess.run(cmd, universal_newlines=text, stdout=subprocess.PIPE)

def assemble(code, fname):
    """ run assembly code through powerpc-eabi-as """
    sFile = fname + '.s'
    oFile = fname + '.o'
    createFile(code, sFile)
    runCmd(['powerpc-eabi-as', sFile, '-o', oFile])

def link(script, outfile, flags=None, entryPoint=None, verbose=False):
    """ run powerpc-eabi-ld with the provided linker script """
    createFile(script, 'linker_script.txt')
    cmd = ['powerpc-eabi-ld', '-o', outfile, '--script', 'linker_script.txt']
    if flags is not None:
        cmd += flags
    if entryPoint is not None:
        cmd += ['-e', entryPoint]
    runCmd(cmd, verbose)

def roundUpToN(x, n):
    """ if x is not divisble by n, round up to nearest multiple of n """
    rounded = n * int(x/n)
    if rounded != x:
        return rounded + n
    return x

def isLineOfCode(line):
    """ return True is string represents a line of assembly code in a object file """
    if ':' in line:
        line = line.split(':')[0]
        return '.' not in line and '>' not in line and '_' not in line
    return False

def getCodeFromObjectFile(filename):
    """ return a dictionary of assembly code from an object file """
    code = {}
    res = runCmd(['powerpc-eabi-objdump', '-D', filename])
    for line in res.stdout.split('\n'):
        if isLineOfCode(line):
            address = int(line.split(':')[0], 16)
            if 'Address' in line:
                print('Warning:', line.split(':')[1])
            else:
                code[address] = int(''.join(line.split()[1:5]), 16)
    return code

def disassembleSections(filename):
    dump = {}
    res = runCmd(['powerpc-eabi-objdump', '-D', filename])
    for section in res.stdout.split('Disassembly of section wiimake_section_')[1:]:
        secNumber = int(section.split(':')[0])
        dump[secNumber] = ''.join(section.split(':')[1:]).strip()
    return dump

def addToDictonary(dict1, dict2):
    """ add one dictionary to another, making sure all duplicate items match """
    for key in dict2.keys():
        if key in dict1.keys():
            if dict1[key] != dict2[key]:
                print(dict1[key])
                print(dict2[key])
                sys.exit("Error: dictionary value mismatch")
        else:
            dict1[key] = dict2[key]

def addOverwrittenInstruction(elfFile, instruction):
    sectionHeaderTableStart = readFile(elfFile, 0x20, 4)
    sectionHeaderSize = readFile(elfFile, 0x2e, 2)
    # the setup code is in the second section, see WiimakePipeline.createBranchingCode()
    sectionNum = 2
    sectionHeaderOffset = sectionHeaderTableStart + sectionNum * sectionHeaderSize
    sectionOffset = readFile(elfFile, sectionHeaderOffset + 0x10, 4)
    sectionSize = readFile(elfFile, sectionHeaderOffset + 0x14, 4)
    instructionOffset = sectionOffset + sectionSize - 0x8 # two lines before end
    if readFile(elfFile, instructionOffset, 4) != 0x60000000:
        print("ERROR: nop not at overwrite address")
    writeFile(elfFile, instruction, instructionOffset)

class ObjectFileSection():
    def __init__(self, name, address=0):
        self.name = name
        self.address = address
        self.size = 0

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

class WiimakePipeline():
    def __init__(self, config):
        self.config = config
        self.sections = []
        self.objects = []
        self.finalCode = {}
        self.codeDump = {}
        self.verbose = False

    def getCode(self):
        return self.finalCode

    def run(self, verbose=False, keepTempFiles=False):
        self.verbose = verbose
        self.createBranchingCode()
        self.compileSourceFiles()
        self.objects += self.config.libraries
        self.cleanUpObjectFiles()
        self.processObjectFileSections()
        self.findRegionsForCode()
        self.linkCode()
        self.dumpCodeToFile()
        if not keepTempFiles:
            self.deleteTemporaryFiles()

    def createBranchingCode(self):
        assemble(PRESERVE_REGISTERS_ASSEMBLY, 'preserve_registers')
        self.sections.append(ObjectFileSection('preserve_registers.o'))
        for i, sym in enumerate(self.config.fixedSymbols):
            injectPointFilename = 'inject_point_{n}'.format(n=i+1)
            entryPointFilename = 'entry_point_{n}'.format(n=i+1)
            assemble(INJECTION_POINT_ASSEMBLY.format(num=i+1), injectPointFilename)
            assemble(BRANCH_TO_ENTRY_POINT_ASSEMBLY.format(num=i+1, symbol=sym),
                     entryPointFilename)
            self.sections.append(ObjectFileSection(entryPointFilename + '.o'))
            injectAddress = int(self.config.fixedSymbols[sym][0], 16)
            self.sections.append(ObjectFileSection(injectPointFilename + '.o',
                                                   injectAddress))

    def compileSourceFiles(self):
        for file in self.config.sourceFiles:
            cmd = ['powerpc-eabi-gcc']
            cmd += self.config.compilerFlags
            cmd += ['-ffunction-sections', '-fdata-sections']
            cmd += ['-I' + x for x in self.config.includePaths]
            cmd += ['-c', self.config.basedir + file]
            cmd += ['-o', Path(file).stem + '.o']
            self.objects += [Path(file).stem + '.o']
            print("== Compiling", file, "==")
            runCmd(cmd, self.verbose)

    def cleanUpObjectFiles(self):
        for ofile in self.objects:
            runCmd(['powerpc-eabi-objcopy', '-R', '.comment', ofile])

    def processObjectFileSections(self):
        self.getSectionNames()
        self.getSectionSizes()
        self.sections = [x for x in self.sections if x.size > 0]

    def getSectionNames(self):
        for ofile in self.objects:
            res = runCmd(['powerpc-eabi-objdump', '-D', ofile])
            tokens = res.stdout.split()
            for token1, token2 in zip(tokens[:-1], tokens[1:]):
                if token1 == 'section':
                    name = ofile + ' (' + token2.split(':')[0] + ')'
                    self.sections.append(ObjectFileSection(name))

    def getSectionSizes(self):
        print("== Calculating Number of Lines of Code ==")
        script = getSizeLinkerScript(self.sections)
        for i in range(len(self.config.fixedSymbols)):
            flags = ['--gc-sections'] + self.config.linkerFlags
            entry = 'inject_point_{n}'.format(n=i+1)
            link(script, 'sizes.out', flags, entry, self.verbose)
            res = runCmd(['powerpc-eabi-readelf', '-s', 'sizes.out'])
            self.getSizesFromSymbolTable(res.stdout.split('\n'))

    def getSizesFromSymbolTable(self, tableDump):
        for line in tableDump:
            if '_sizeof_section_' in line:
                ndx = line.find(":")
                size = int(line[(ndx+2):(ndx+10)], 16)
                size = roundUpToN(size, 4)
                ndx = line.find("_sizeof_section_")
                sectionId = int(line[(ndx+16):])
                sectionSize = max(size, self.sections[sectionId - 1].size)
                self.sections[sectionId - 1].size = sectionSize

    def findRegionsForCode(self):
        print("== Finding Regions to Inject Code Into ==")
        regions = copy.deepcopy(self.config.memoryRegions)
        sections = self.sections.copy()
        sections = [x for x in sections if x.address == 0]
        sections.sort(key=lambda x: x.size, reverse=True)
        for sec in sections:
            regions.sort(key=lambda x: x.size, reverse=True)
            sec.address = regions[0].start # put at beginning of largest region
            regions[0].start += sec.size + 4 # update the available space
            if regions[0].start > regions[0].end: # check if space ran out
                msg = 'ERROR: failed to place code in available regions.\n'
                msg += 'lines of code to inject: {cLines}\n'
                msg += 'memory regions total capacity: {mLines}\n'
                codeSize = sum([x.size for x in self.sections])
                memSize = sum([r.size for r in self.config.memoryRegions])
                msg = msg.format(cLines=codeSize//4, mLines=memSize//4)
                sys.exit(msg)

    def linkCode(self):
        print("== Linking Final Code ==")
        script = getFinalLinkerScript(self.sections)
        for i, sym in enumerate(self.config.fixedSymbols):
            flags = ['--gc-sections'] + self.config.linkerFlags
            entry = 'inject_point_{n}'.format(n=i+1)
            link(script, 'final.out', flags, entry, self.verbose)
            overwrittenInstruction = int(self.config.fixedSymbols[sym][1], 16)
            addOverwrittenInstruction('final.out', overwrittenInstruction)
            addToDictonary(self.codeDump, disassembleSections('final.out'))
            addToDictonary(self.finalCode, getCodeFromObjectFile('final.out'))

    def dumpCodeToFile(self):
        txtDump = []
        template = '\n\nDisassembly of section wiimake_section_{num}\n\n{code}'
        for num, code in sorted(self.codeDump.items()):
            txtDump.append(template.format(num=num, code=code))
        createFile(''.join(txtDump), 'injected_code.txt')

    def deleteTemporaryFiles(self):
        os.remove('linker_script.txt')
        os.remove('final.out')
        os.remove('sizes.out')
        os.remove('preserve_registers.s')
        os.remove('preserve_registers.o')
        for i in range(len(self.config.fixedSymbols)):
            os.remove('inject_point_{n}.s'.format(n=i+1))
            os.remove('inject_point_{n}.o'.format(n=i+1))
            os.remove('entry_point_{n}.s'.format(n=i+1))
            os.remove('entry_point_{n}.o'.format(n=i+1))
        for file in self.config.sourceFiles:
            os.remove(Path(file).stem + '.o')



import argparse
import sys
from wiimake.iso import Iso

# description for wiimake-isotool
DESCRIPTION = '''
wiimake-isotool is useful for operating on an iso file directly. The --read
flag allows the user to read any RAM address from the iso, which is useful when
trying to find the original instruction that is being overwritten at the
injection point. The --checksum option allows for simple verification of the
iso version. There is also --save and --load which allow for easy distribution
of mods.
'''

def singleTrue(iterable):
    """ returns True if input contains exactly one True argument

    Credit to https://stackoverflow.com/a/16801605
    """
    iterator = iter(iterable)
    hasTrue = any(iterator) # stops after true found (or end reached)
    hasAnotherTrue = any(iterator)
    return hasTrue and not hasAnotherTrue

def commandLineArgError(parser, msg):
    """ print usage message before exiting with an error """
    parser.print_usage()
    sys.exit("Error:" + msg)

def getAddress(addr):
    """ convert a string representing a file position or address to an integer """
    try:
        address = int(addr, 0)
    except ValueError:
        sys.exit("Error: if address if a hex value, it must start with 0x")
    return address

def main():
    """main function for wiimake-isotool"""

    # parse command line arguments
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--version', action='version', version='%(prog)s 1.0')
    parser.add_argument("file", help="iso file for a Gamecube/Wii game")
    parser.add_argument("--save", metavar="file",
                        help="creates a save state for the iso file")
    parser.add_argument("--load", metavar="file",
                        help="loads a save state for the iso file")
    parser.add_argument("--read", metavar="address/position",
                        help='''reads the value at the given position in the
                                file. If the value could be a memory address in
                                the DOL, the corresponding value at that memory
                                address is read instead''')
    parser.add_argument("--checksum", action="store_true",
                        help="computes an md5 checksum of the file")
    parser.add_argument("--print-dol-table", action="store_true",
                        help="prints the DOL table of the iso")
    parser.add_argument("--zero-out", nargs=2, metavar=("address", "size"),
                        help='''zereos out a section of memory, first argument
                                is starting address, second argument is number
                                of bytes to zero out. This is useful when
                                exploring the effects of overwriting certain
                                regions of memory''')
    args = parser.parse_args()

    # process command line arguments, dispatch to correct function
    optionalArgs = [args.read, args.save, args.load, args.checksum,
                    args.print_dol_table, args.zero_out]
    if not singleTrue(optionalArgs):
        commandLineArgError(parser, "must provide exactly one option")
    elif args.read is not None:
        address = getAddress(args.read)
        print(hex(Iso(args.file).read(address, True)))
    elif args.save is not None:
        print("saving file state...")
        Iso(args.file).save(args.save)
        print("done!")
    elif args.load is not None:
        print("loading file state...")
        Iso(args.file).load(args.load)
        print("done!")
    elif args.zero_out is not None:
        code = {}
        start = getAddress(args.zero_out[0])
        end = getAddress(args.zero_out[1])
        for addr in range(start, end, 4):
            code[addr] = 0
        Iso(args.file).bulkWrite(code)
    elif args.print_dol_table:
        print(Iso(args.file).dolTable)
    elif args.checksum:
        iso = Iso(args.file)
        print(iso.md5(), '', iso.file)



import re
from setuptools import setup

version = re.search(
    '^__version__\s*=\s*\'(.*)\'',
    open('wiimake/_version.py').read(),
    re.M
    ).group(1)

with open("README.rst", "rb") as f:
    long_descr = f.read().decode("utf-8")

setup(
    name='wiimake',
    version=version,
    packages=['wiimake'],
    description='Command line tool for building Gamecube/Wii mods',
    long_description=long_descr,
    author='Tom Sherman',
    author_email='tomsherman159@gmail.com',
    url='https://github.com/sherman5/wiimake',
    entry_points={
        'console_scripts': [
            'wiimake=wiimake.wiimake:main',
            'wiimake-isotool=wiimake.wiimake_isotool:main'
        ]
    }
)



PRESERVE_REGISTERS_ASSEMBLY = '''
.global backup_reg
backup_reg:
    subi 1,1,128
    stmw 3,12(1)
    stw 0,8(1)
    mfcr 0
    stw 0,4(1)
    mfctr 0
    stw 0,0(1)
    blr

.global restore_reg
restore_reg:
    lwz 0,0(1)
    mtctr 0
    lwz 0,4(1)
    mtcr 0
    lwz 0,8(1)
    lmw 3,12(1)
    addi 1,1,128
    blr
'''

INJECTION_POINT_ASSEMBLY = '''
.global inject_point_{num}
inject_point_{num}:
    b entry_point_{num}
'''

BRANCH_TO_ENTRY_POINT_ASSEMBLY = '''
.global entry_point_{num}
entry_point_{num}:
    subi 1,1,4
    stw  0,4(1)
    mflr 0
    bl backup_reg
    bl {symbol}
    bl restore_reg
    mtlr 0
    lwz 0,4(1)
    addi 1,1,4
    nop
    b inject_point_{num} + 0x04
'''